from .version import (
    __author__, __version__, author_info, package_info,
    package_license, project_home, team_email, version_info,
)

from .protocol import *
from .context import *
from .setup import setup

__all__ = (
    "__author__",
    "__version__",
    "author_info",
    "package_info",
    "package_license",
    "project_home",
    "team_email",
    "version_info",

    # Clients
    "TCPClient",
    "UDPClient",
    "PickleClient",
    "Meter",
    "Counter",
    "Timer",
    "setup",
)



import asyncio
import io
import logging
import socket

from aiocarbon.metric import Metric
from aiocarbon.storage.base import BaseStorage
from .base import BaseClient


log = logging.getLogger(__name__)


async def resolver(host, port, loop):
    addr_info = await loop.getaddrinfo(host, port)

    family, _, _, _, host_info = next(
        filter(
            lambda x: x[1] == socket.SOCK_DGRAM,
            addr_info
        )
    )

    return family, host_info[0], port


class AsyncUDPSocket:
    __slots__ = (
        '__loop', '__sock', '__address',
        '__futures', '__closed', '__writer_added',
    )

    @staticmethod
    def create_future(loop):
        if hasattr(loop, 'create_future'):
            return loop.create_future()
        else:
            return asyncio.Future(loop=loop)

    def __init__(self, loop=None):
        self.__loop = asyncio.get_event_loop() if loop is None else loop
        self.__sock = None      # type: socket.socket

        self.__futures = list()
        self.__closed = False
        self.__writer_added = False

    async def sendto(self, data, host, port):
        family, host, port = await resolver(host, port, self.__loop)

        if self.__sock is None:
            self.__sock = socket.socket(family, socket.SOCK_DGRAM)
            self.__sock.setblocking(False)

        data = data if isinstance(data, bytes) else str(data).encode('utf-8')
        future = self.create_future(self.__loop)

        destination = (data, host, port)
        self.__futures.append((destination, future))

        if not self.__writer_added:
            self.__loop.add_writer(self.__sock.fileno(), self.__sender)
            self.__writer_added = True

        return await future

    def __sender(self):
        if not self.__futures:
            self.__loop.remove_writer(self.__sock.fileno())
            self.__writer_added = False
            return

        destination, future = self.__futures[0]
        data, host, port = destination

        try:
            self.__sock.sendto(data, (host, port))
        except (BlockingIOError, InterruptedError):
            return
        except Exception as exc:
            self.__abort(exc)
        else:
            self.__futures.pop(0)
            future.set_result(True)

    def __abort(self, exc):
        for future in (f for _, f in self.__futures if not f.done()):
            future.set_exception(exc)

        self.close()

    @property
    def is_closed(self):
        return self.__closed

    def close(self):
        if self.__closed:
            raise RuntimeError("Socket already closed")

        self.__closed = True
        self.__loop.remove_writer(self.__sock.fileno())
        self.__sock.close()

        for future in (f for _, f in self.__futures if not f.done()):
            future.set_exception(ConnectionError("Connection closed"))


class UDPClient(BaseClient):
    SENDING_THRESHOLD = 1200

    def __init__(self, host: str, port: int = 2003, namespace: str = None,
                 loop: asyncio.AbstractEventLoop = None,
                 storage: BaseStorage = None):

        super().__init__(host, port,
                         namespace=namespace,
                         loop=loop,
                         storage=storage)

        self._socket = AsyncUDPSocket(loop=self.loop)

    async def _send_part(self, data):
        await self._socket.sendto(data, self._host, self._port)
        log.debug('%d bytes were sent', len(data))

    async def send(self):
        async with self.lock:
            with io.BytesIO() as buffer:
                for metric in self._storage:
                    buffer.write(self.format_metric(metric))

                    if buffer.tell() > self.SENDING_THRESHOLD:
                        await self._send_part(buffer.getvalue())

                        buffer.seek(0)
                        buffer.truncate(0)

                if buffer.tell():
                    await self._send_part(buffer.getvalue())

    def format_metric(self, metric: Metric) -> bytes:
        if isinstance(metric.value, float):
            value = "%.12f" % metric.value
        else:
            value = metric.value

        return (
            "%s %s %s\n" % (
                self.format_metric_name(metric), value, metric.timestamp,
            )
        ).encode()



import time
from collections import Counter
from typing import AsyncIterable

from immutables import Map

from aiocarbon.storage.base import BaseStorage
from aiocarbon.metric import Metric


class TotalStorage(BaseStorage):

    """ Adds up values of the same metrics with the same timestamps """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metrics = Map()
        self._storage_class = Counter

    def __iter__(self) -> AsyncIterable[Metric]:
        current_time = int(time.time())

        for name, metrics in self._metrics.items():  # type: Counter
            returning = list()

            while metrics:
                ts, value = metrics.popitem()

                if ts >= current_time:
                    returning.append((ts, value))
                    continue

                yield Metric(name=name, timestamp=ts, value=value)

            for ts, value in returning:
                metrics[ts] += value



from collections import deque, namedtuple

from aiocarbon.metric import Metric
from aiocarbon.storage.base import Operations


BufferedMetric = namedtuple('BufferedMetric', ['metric', 'operation'])


class BufferClient:

    """
    Special case client for metrics buffering before initialization of
    proper carbon client completed
    """

    MAX_BUFFER_LEN = 10000

    def __init__(self):
        self.metrics_buffer = deque([], maxlen=self.MAX_BUFFER_LEN)

    def add(self, metric, operation):
        self.metrics_buffer.append(BufferedMetric(metric, operation))



import abc
from typing import AsyncIterable

from aiocarbon.metric import Metric


class Operations:
    @staticmethod
    def add(store, metric: Metric):
        store[metric.timestamp] += metric.value

    @staticmethod
    def avg(store, metric: Metric):
        store[metric.timestamp] = (store[metric.timestamp] + metric.value) / 2


class BaseStorage:

    def __init__(self):
        self._metrics = None
        self._storage_class = None

    @abc.abstractmethod
    def __iter__(self) -> AsyncIterable[Metric]:
        raise NotImplementedError

    def _get_metric(self, name):
        if name not in self._metrics:
            self._metrics = self._metrics.set(name, self._storage_class())

        return self._metrics[name]

    def add(self, metric: Metric, operation=Operations.add):
        operation(self._get_metric(metric.name), metric)

    def poll(self):
        return any(self._metrics.values())



author_info = (
    ('Dmitry Orlov', 'me@mosquito.su'),
)

project_home = 'http://github.com/mosquito/aiocarbon'

package_info = "Asynchronous client for carbon."
package_license = "Apache 2"

team_email = 'me@mosquito.su'

version_info = (0, 15, 0)

__author__ = ", ".join("{} <{}>".format(*info) for info in author_info)
__version__ = ".".join(map(str, version_info))



import asyncio
import logging
import pickle
import struct
from typing import Tuple, List

from aiocarbon.metric import Metric
from .base import BaseClient

log = logging.getLogger(__name__)


class PickleClient(BaseClient):
    CHUNK_SIZE = 512

    async def __sender(self, data: List):
        payload = pickle.dumps(data, protocol=2)
        header = struct.pack("!L", len(payload))

        while True:
            try:
                reader, writer = await asyncio.open_connection(
                    self._host, self._port, loop=self.loop
                )

                writer.write(header)
                await writer.drain()
                writer.write(payload)
                await writer.drain()

                writer.close()
                reader.feed_eof()
            except:
                log.exception("Failed to send metrics")
            else:
                break

    async def send(self):
        async with self.lock:
            data = []

            for idx, metric in enumerate(self._storage):
                data.append(self.format_metric(metric))

                if idx % self.CHUNK_SIZE == 0:
                    await self.__sender(data)
                    data.clear()

            if data:
                await self.__sender(data)

    def format_metric(self, metric: Metric) -> Tuple[str, Tuple[float, int]]:
        return (
            self.format_metric_name(metric),
            (metric.timestamp, metric.value)
        )



import threading
from typing import ClassVar

import time

from .protocol.base import BaseClient
from .protocol.buffer import BufferClient
from .storage.base import Operations
from .metric import Metric


class Meter:
    __slots__ = "_name", "value", "timestamp", "suffix"

    TLS = threading.local()
    TLS.client = BufferClient()

    def __init__(self, name, value=None, timestamp=None, suffix=None):
        self._name = name
        self.value = value
        self.timestamp = timestamp or int(time.time())
        self.suffix = suffix

    def send(self, operation: ClassVar[Operations]=Operations.add):
        if self.value and self.timestamp:
            if self.suffix:
                self._name = ".".join((self._name, self.suffix))

            self.TLS.client.add(
                Metric(self._name, self.value, self.timestamp),
                operation=operation
            )


class Counter(Meter):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self._name = "%s.%s" % (self._name.rstrip("."), 'fail')
        else:
            self._name = "%s.%s" % (self._name.rstrip("."), 'ok')

        self.value = 1
        self.send(Operations.add)


class Timer(Meter):
    __slots__ = "_start_time",

    def __enter__(self):
        self._start_time = time.monotonic()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.value = time.monotonic() - self._start_time

        if exc_type:
            self._name = "%s.%s" % (self._name.rstrip("."), 'fail')
        else:
            self._name = "%s.%s" % (self._name.rstrip("."), 'ok')

        self.send(Operations.avg)


def set_client(client: BaseClient):
    buffer_client, Meter.TLS.client = Meter.TLS.client, client

    if hasattr(buffer_client, 'metrics_buffer'):
        while len(buffer_client.metrics_buffer):
            metric, operation = buffer_client.metrics_buffer.popleft()
            client.add(metric, operation)


__all__ = "Meter", "Counter", "Timer", "set_client",



import asyncio
import logging

from aiocarbon.metric import Metric
from .base import BaseClient


log = logging.getLogger(__name__)


class TCPClient(BaseClient):
    CHUNK_SIZE = 512

    async def send(self):
        async with self.lock:
            reader, writer = await asyncio.open_connection(
                self._host, self._port, loop=self.loop
            )

            idx = 0
            for idx, metric in enumerate(self._storage):
                value = self.format_metric(metric)
                writer.write(value)

                if idx % self.CHUNK_SIZE == 0:
                    await writer.drain()

            log.debug('%d metric(s) were sent', idx + 1)

            await writer.drain()
            writer.close()
            reader.feed_eof()

    def format_metric(self, metric: Metric) -> bytes:
        if isinstance(metric.value, float):
            value = "%.12f" % metric.value
        else:
            value = metric.value

        return (
            "%s %s %s\n" % (
                self.format_metric_name(metric), value, metric.timestamp,
            )
        ).encode()



import abc
import asyncio
import logging
import re

from aiocarbon.metric import Metric
from aiocarbon.storage.base import BaseStorage
from aiocarbon.storage import TotalStorage


log = logging.getLogger(__name__)

CARBON_NS_INVALID_CHARS = re.compile(r'([^\w\d\.\-]+|(\.){2,})')


def strip_carbon_ns(string):
    return CARBON_NS_INVALID_CHARS.sub('_', string).strip('_').lower()


class BaseClient:
    SEND_PERIOD = 1

    def __init__(self, host: str, port: int = 2003,
                 loop: asyncio.AbstractEventLoop = None,
                 storage: BaseStorage = None,
                 namespace: str=None):
        self.loop = loop or asyncio.get_event_loop()
        self.namespace = namespace if namespace is not None else 'aiocarbon'
        self._host = host
        self._port = port
        self._storage = storage or TotalStorage()
        self.lock = asyncio.Lock(loop=self.loop)

    @property
    def namespace(self):
        return self._ns

    @namespace.setter
    def namespace(self, value):
        self._ns = strip_carbon_ns(value)

    async def run(self):
        while True:
            try:
                await self.send()
            except:
                log.exception("Exception while sending metrics")
            finally:
                await asyncio.sleep(self.SEND_PERIOD, loop=self.loop)

    def format_metric_name(self, metric: Metric):
        if self._ns:
            return ".".join((self._ns, metric.name))
        else:
            return metric.name

    @abc.abstractmethod
    def send(self):
        raise NotImplementedError

    @abc.abstractmethod
    def format_metric(self, metric: Metric):
        raise NotImplementedError

    def add(self, metric: Metric, operation=None):
        kwargs = {}
        if operation is not None:
            kwargs['operation'] = operation
        return self._storage.add(metric, **kwargs)



import asyncio
from typing import Type

from .protocol.base import BaseClient
from .protocol import PickleClient
from .context import set_client


def setup(host: str, port: int, client_class: Type[BaseClient]=PickleClient,
          namespace: str="carbon", loop: asyncio.AbstractEventLoop=None):

    loop = loop or asyncio.get_event_loop()
    client = client_class(host=host, port=port, namespace=namespace, loop=loop)
    set_client(client)

    return client



from setuptools import setup, find_packages
from os import path

from importlib.machinery import SourceFileLoader


module = SourceFileLoader(
    "version", path.join("aiocarbon", "version.py")
).load_module()


setup(
    name='aiocarbon',
    version=module.__version__,
    packages=find_packages(exclude=['tests', 'example', 'env']),
    license=module.package_license,
    description=module.package_info,
    long_description=open("README.rst").read(),
    url=module.project_home,
    author=module.__author__,
    author_email=module.team_email,
    provides=["aiocarbon"],
    python_requires=">3.5.*, <4",
    keywords=["aio", "python", "asyncio", "carbon", "graphite", "client"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: System',
        'Topic :: System :: Operating System',
    ],
    install_requires=[
        "immutables",
    ]
)



import time
from typing import AsyncIterable
import logging

from immutables import Map

from aiocarbon.storage.base import BaseStorage
from aiocarbon.metric import Metric


log = logging.getLogger(__name__)


class RawStorage(BaseStorage):

    """ Saves metrics as is """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._metrics = Map()
        self._storage_class = list

    def __iter__(self) -> AsyncIterable[Metric]:
        current_time = int(time.time())

        for name, metrics in self._metrics.items():  # type: Counter
            returning = list()

            for ts, value in metrics:

                if ts >= current_time:
                    returning.append((ts, value))
                    continue

                yield Metric(name=name, timestamp=ts, value=value)

            metrics.clear()

            if not returning:
                continue

            metrics.extend(returning)
            log.info("%d metric(s) of %r weren't sent yet because they're "
                     "newer than the current timestamp %d",
                     len(returning), name, current_time)

    def add(self, metric: Metric, operation=None):
        self._get_metric(metric.name).append((metric.timestamp, metric.value))



from collections import namedtuple

import time


class Metric(namedtuple("Metric", ("name", "value", "timestamp"))):
    def __new__(cls, name, value, timestamp=None):
        timestamp = int(timestamp or time.time())

        if not isinstance(value, (int, float)):
            raise TypeError("Value should be int or float not %r" % type(value))

        return super().__new__(cls, name=name, value=value, timestamp=timestamp)

    def __str__(self):
        return "{0.name} {0.value} {0.timestamp}".format(self)

    def __repr__(self):
        return "<Metric: {0.name} {0.value} {0.timestamp}>".format(self)
